
#include <stdio.h>
#include <math.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include "input.h"
#include "pnginput.h"
#include "xoutput.h"
#include "tools.h"
#include "leetft.h"
#include "fifo.h"

static double intensity_to_voltage_srgb( double val )
{
    /* Handle invalid values before doing a gamma transform. */
    if( val < 0.0 ) return 0.0;
    if( val > 1.0 ) return 1.0;

    /* sRGB uses an exponent of 2.4 in its nonstandard curve, but it
     *      * still advertises a display gamma of 2.2.
     *           */
    if( val <= 0.0031308 ) {
        return 12.92 * val;
    }
    return ( ( 1.055 * pow( val, 1.0 / 2.4 ) ) - 0.055 );
}


#define MAX_HEADLINES 50
static char *headlines[ MAX_HEADLINES ];
static int headlines_init = 0;
static int headlines_cur = 0;
static fifo_t *hf = 0;
static char *default_headline = "NO HEADLINE AVAILABLE AT THIS TIME.  ";
static char *freestring = 0;
int last_headline = 0;

const char *get_next_headline( void )
{
    int i;

    if( !headlines_init ) {
        hf = fifo_new( "/tmp/newsfifo" );
        if( !hf ) {
            fprintf( stderr, "no fifo\n" );
            exit( 1 );
        }
        for( i = 0; i < MAX_HEADLINES; i++ ) headlines[ i ] = 0;
        headlines_init = 1;
    }

    if( freestring ) {
        free( freestring );
        freestring = 0;
        headlines[ last_headline ] = 0;
    }

    if( !headlines[ headlines_cur ] ) {
        const char *nextline = fifo_next_line( hf );

        if( nextline ) {
            int needsize = strlen( default_headline );
            if( strlen( nextline ) < needsize ) {
                headlines[ headlines_cur ] = (char *) malloc( needsize );
                strcpy( headlines[ headlines_cur ], nextline );
                for( i = strlen( nextline ) + 1; i < needsize; i++ ) {
                    headlines[ headlines_cur ][ i - 1 ] = '.';
                }
                headlines[ headlines_cur ][ needsize - 1 ] = '\0';
            } else {
                headlines[ headlines_cur ] = strdup( nextline );
            }
            fifo_done_line( hf );
        } else {
            return default_headline;
        }
    }

    freestring = headlines[ headlines_cur ];
    headlines_cur = (headlines_cur + 1) % MAX_HEADLINES;
    return freestring;
}

int main( int argc, char **argv )
{
    output_api_t *output = xoutput_get_output();
    input_t *in = input_new();
    int width, height, stride;
    ft_font_t *font;
    ft_string_t *h1s;
    ft_string_t *h2s;
    double pixel_aspect;
    int i, j;
    int pos = 0;
    int h1w = 0;
    int h2w = 0;
    int curh = 0;
    unsigned char *prerender;

    // int accel_max = 280000;
    int accel_max = 100000;
    int accel_step = 4000;
    int accel = 0;
    // int accel = 4096;
    int paused = 1;

    width = 640;
    height = 60;
    output->init( width, height );
    stride = output->get_output_stride();

    prerender = (unsigned char *) malloc( width * height );
    if( !prerender ) {
        fprintf( stderr, "you suck mcbain\n" );
        return 1;
    }

    pixel_aspect = 1.0;
    font = ft_font_new( "FreeSansBold.ttf", 30, pixel_aspect );
    if( !font ) {
        fprintf( stderr, "can't open font?\n" );
        return 1;
    }

    h1s = ft_string_new( font );
    h2s = ft_string_new( font );

    ft_string_set_text( h1s, get_next_headline(), pos );
    ft_string_set_text( h2s, get_next_headline(), pos );

    output->lock_output_buffer();
    memset( output->get_output_buffer(), 0, stride * height );
    output->unlock_output_buffer();
    output->show_frame();

    for(;;) {
        int curw = ft_string_get_width( h1s );

        output->poll_events( in );
        if( input_quit_requested( in ) ) break;
        if( input_pause_requested( in ) ) paused = !paused;
        input_next_frame( in );

        if( !paused ) {
            pos += accel;
            if( accel < accel_max ) {
                accel += accel_step;
            }
        } else {
            pos += accel;
            if( accel > 0 ) {
                accel -= accel_step;
            }
        }

        if( pos > (curw << 16) ) {
            ft_string_t *tmph;
            pos -= (curw << 16);
            tmph = h2s;
            h2s = h1s;
            h1s = tmph;
            ft_string_set_text( h2s, get_next_headline(), pos );
        }

        /* pre-render */
        memset( prerender, 0, width * height );
        h1w = ft_string_get_width( h1s ) - ( pos >> 16 );
        if( h1w > width ) h1w = width;
        for( i = 0; i < ft_string_get_height( h1s ); i++ ) {
            unsigned char *curdest = prerender + (i * width);
            unsigned char *cursrc = ft_string_get_buffer( h1s ) + (i * ft_string_get_stride( h1s ));
            a8_subpix_blit_scanline( curdest, cursrc + (pos >> 16), pos & 0xffff, h1w );
        }
        if( h1w < width ) {
            h2w = width - h1w;
            if( h2w > ft_string_get_width( h2s ) ) {
                h2w = ft_string_get_width( h2s );
            }
            for( i = 0; i < ft_string_get_height( h2s ); i++ ) {
                unsigned char *curdest = prerender + (i * width) + h1w;
                unsigned char *cursrc = ft_string_get_buffer( h2s ) + (i * ft_string_get_stride( h2s ));
                a8_subpix_blit_scanline( curdest, cursrc, pos, h2w );
            }
        }

        output->lock_output_buffer();
        memset( output->get_output_buffer(), 0, stride * height );

        for( i = 0; i < height; i++ ) {
            unsigned char *curscanline = output->get_output_buffer() + (stride * 5) + (stride*i);
            unsigned char *fdata = prerender + (i * width);
            for( j = 0; j < width; j++ ) {
                if( fdata[ j ] ) {
                    (curscanline + (j*4))[ 0 ] = (int)((255.0 * intensity_to_voltage_srgb( ((double) fdata[ j ]) / 255.0 ))+0.5);
                    // (curscanline + (j*4))[ 0 ] = fdata[ j ];
                    (curscanline + (j*4))[ 1 ] = (curscanline + (j*4))[ 0 ];
                    (curscanline + (j*4))[ 2 ] = (curscanline + (j*4))[ 0 ];
                    (curscanline + (j*4))[ 3 ] = 0xff;
                }
            }
        }
        output->unlock_output_buffer();
        output->show_frame();
        usleep( 10000 );
    }

    output->shutdown();
    return 0;
}

